@@ -1,7 +1,7 @@
 
 /*
  * BZip3 - A spiritual successor to BZip2.
- * Copyright (C) 2022-2023 Kamila Szewczyk
+ * Copyright (C) 2022-2024 Kamila Szewczyk
  *
  * This program is free software: you can redistribute it and/or modify it
  * under the terms of the GNU Lesser General Public License as published by the Free
@@ -20,18 +20,13 @@
 #include <ctype.h>
 #include <errno.h>
 #include <inttypes.h>
+#include <limits.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <sys/stat.h>
 #include <unistd.h>
 
-#ifdef HAVE_GETOPT_LONG
-    #include <getopt.h>
-#else
-    #include "getopt-shim.h"
-#endif
-
 #if defined __MSVCRT__
     #include <fcntl.h>
     #include <io.h>
@@ -39,6 +34,7 @@
 
 #include "common.h"
 #include "libbz3.h"
+#include "yarg.h"
 
 #define MODE_DECODE 0
 #define MODE_ENCODE 1
@@ -59,10 +55,11 @@ static void help() {
             "Operations:\n"
             "  -e/-z, --encode   compress data (default)\n"
             "  -d, --decode      decompress data\n"
-			"  -r, --recover     attempt at recovering corrupted data\n"
+            "  -r, --recover     attempt at recovering corrupted data\n"
             "  -t, --test        verify validity of compressed data\n"
             "  -h, --help        display an usage overview\n"
             "  -f, --force       force overwriting output if it already exists\n"
+            "      --rm          remove input files after successful (de)compression\n"
             "  -k, --keep        keep (don't delete) input files (default)\n"
             "  -v, --verbose     verbose mode (display more information)\n"
             "  -V, --version     display version information\n"
@@ -78,8 +75,7 @@ static void help() {
 }
 
 static void xwrite(const void * data, size_t size, size_t len, FILE * des) {
-    if (len == 0 || size == 0)
-        return;
+    if (len == 0 || size == 0) return;
     if (fwrite(data, size, len, des) != len) {
         fprintf(stderr, "Write error: %s\n", strerror(errno));
         exit(1);
@@ -145,6 +141,19 @@ static void close_out_file(FILE * des) {
     }
 }
 
+static void remove_in_file(char * file_name, FILE * output_des) {
+    if (file_name == NULL) {
+        return;
+    }
+    if (output_des == stdout) {
+        return;
+    }
+    if (remove(file_name)) {
+        fprintf(stderr, "Error: failed to remove input file `%s': %s\n", file_name, strerror(errno));
+        exit(1);
+    }
+}
+
 static int process(FILE * input_des, FILE * output_des, int mode, int block_size, int workers, int verbose,
                    char * file_name) {
     uint64_t bytes_read = 0, bytes_written = 0;
@@ -169,7 +178,7 @@ static int process(FILE * input_des, FILE * output_des, int mode, int block_size
 
             bytes_written += 9;
             break;
-		case MODE_RECOVER:
+        case MODE_RECOVER:
         case MODE_DECODE:
         case MODE_TEST: {
             char signature[5];
@@ -187,12 +196,12 @@ static int process(FILE * input_des, FILE * output_des, int mode, int block_size
                 fprintf(stderr,
                         "The input file is corrupted. Reason: Invalid block "
                         "size in the header.\n");
-				if(mode == MODE_RECOVER) {
-					fprintf(stderr, "Recovery mode: Proceeding.\n");
-					block_size = MiB(511);
-				} else {
-					return 1;
-				}
+                if (mode == MODE_RECOVER) {
+                    fprintf(stderr, "Recovery mode: Proceeding.\n");
+                    block_size = MiB(511);
+                } else {
+                    return 1;
+                }
             }
 
             bytes_read += 9;
@@ -215,7 +224,8 @@ static int process(FILE * input_des, FILE * output_des, int mode, int block_size
             return 1;
         }
 
-        u8 * buffer = malloc(bz3_bound(block_size));
+        size_t buffer_size = bz3_bound(block_size);
+        u8 * buffer = malloc(buffer_size);
 
         if (!buffer) {
             fprintf(stderr, "Failed to allocate memory.\n");
@@ -258,7 +268,7 @@ static int process(FILE * input_des, FILE * output_des, int mode, int block_size
                 }
                 xread_noeof(buffer, 1, new_size, input_des);
                 bytes_read += 8 + new_size;
-                if (bz3_decode_block(state, buffer, new_size, old_size) == -1) {
+                if (bz3_decode_block(state, buffer, buffer_size, new_size, old_size) == -1) {
                     fprintf(stderr, "Failed to decode a block: %s\n", bz3_strerror(state));
                     return 1;
                 }
@@ -280,7 +290,7 @@ static int process(FILE * input_des, FILE * output_des, int mode, int block_size
                 }
                 xread_noeof(buffer, 1, new_size, input_des);
                 bytes_read += 8 + new_size;
-                if (bz3_decode_block(state, buffer, new_size, old_size) == -1) {
+                if (bz3_decode_block(state, buffer, buffer_size, new_size, old_size) == -1) {
                     fprintf(stderr, "Writing invalid block: %s\n", bz3_strerror(state));
                 }
                 xwrite(buffer, old_size, 1, output_des);
@@ -301,7 +311,7 @@ static int process(FILE * input_des, FILE * output_des, int mode, int block_size
                 xread_noeof(buffer, 1, new_size, input_des);
                 bytes_read += 8 + new_size;
                 bytes_written += old_size;
-                if (bz3_decode_block(state, buffer, new_size, old_size) == -1) {
+                if (bz3_decode_block(state, buffer, buffer_size, new_size, old_size) == -1) {
                     fprintf(stderr, "Failed to decode a block: %s\n", bz3_strerror(state));
                     return 1;
                 }
@@ -321,6 +331,7 @@ static int process(FILE * input_des, FILE * output_des, int mode, int block_size
         struct bz3_state * states[workers];
         u8 * buffers[workers];
         s32 sizes[workers];
+        size_t buffer_sizes[workers];
         s32 old_sizes[workers];
         for (s32 i = 0; i < workers; i++) {
             states[i] = bz3_new(block_size);
@@ -328,7 +339,9 @@ static int process(FILE * input_des, FILE * output_des, int mode, int block_size
                 fprintf(stderr, "Failed to create a block encoder state.\n");
                 return 1;
             }
-            buffers[i] = malloc(block_size + block_size / 50 + 32);
+            size_t buffer_size = bz3_bound(block_size);
+            buffer_sizes[i] = buffer_size;
+            buffers[i] = malloc(buffer_size);
             if (!buffers[i]) {
                 fprintf(stderr, "Failed to allocate memory.\n");
                 return 1;
@@ -379,7 +392,7 @@ static int process(FILE * input_des, FILE * output_des, int mode, int block_size
                     xread_noeof(buffers[i], 1, sizes[i], input_des);
                     bytes_read += 8 + sizes[i];
                 }
-                bz3_decode_blocks(states, buffers, sizes, old_sizes, i);
+                bz3_decode_blocks(states, buffers, buffer_sizes, sizes, old_sizes, i);
                 for (s32 j = 0; j < i; j++) {
                     if (bz3_last_error(states[j]) != BZ3_OK) {
                         fprintf(stderr, "Failed to decode data: %s\n", bz3_strerror(states[j]));
@@ -407,7 +420,7 @@ static int process(FILE * input_des, FILE * output_des, int mode, int block_size
                     xread_noeof(buffers[i], 1, sizes[i], input_des);
                     bytes_read += 8 + sizes[i];
                 }
-                bz3_decode_blocks(states, buffers, sizes, old_sizes, i);
+                bz3_decode_blocks(states, buffers, buffer_sizes, sizes, old_sizes, i);
                 for (s32 j = 0; j < i; j++) {
                     if (bz3_last_error(states[j]) != BZ3_OK) {
                         fprintf(stderr, "Writing invalid block: %s\n", bz3_strerror(states[j]));
@@ -435,7 +448,7 @@ static int process(FILE * input_des, FILE * output_des, int mode, int block_size
                     bytes_read += 8 + sizes[i];
                     bytes_written += old_sizes[i];
                 }
-                bz3_decode_blocks(states, buffers, sizes, old_sizes, i);
+                bz3_decode_blocks(states, buffers, buffer_sizes, sizes, old_sizes, i);
                 for (s32 j = 0; j < i; j++) {
                     if (bz3_last_error(states[j]) != BZ3_OK) {
                         fprintf(stderr, "Failed to decode data: %s\n", bz3_strerror(states[j]));
@@ -538,90 +551,73 @@ int main(int argc, char * argv[]) {
     int force = 0;
 
     // command line arguments
-    int force_stdstreams = 0, workers = 0, batch = 0, verbose = 0;
+    int force_stdstreams = 0, workers = 0, batch = 0, verbose = 0, remove_input_file = 0;
 
     // the block size
     u32 block_size = MiB(16);
 
+    enum { RM_OPTION = CHAR_MAX + 1 };
+
+    yarg_options opt[] = {
+        {       'e', no_argument,       "encode" },
+        {       'z', no_argument,       "encode" }, /* alias */
+        {       'd', no_argument,       "decode" },
+        {       't', no_argument,       "test" },
+        {       'c', no_argument,       "stdout" },
+        {       'f', no_argument,       "force" },
+        {       'r', no_argument,       "recover" },
+        {       'h', no_argument,       "help" },
+        { RM_OPTION, no_argument,       "rm" },
+        {       'k', no_argument,       "keep" },
+        {       'V', no_argument,       "version" },
+        {       'v', no_argument,       "verbose" },
+        {       'b', required_argument, "block" },
+        {       'B', no_argument,       "batch" },
 #ifdef PTHREAD
-    const char * short_options = "Bb:cdefhj:krtvVz";
-#else
-    const char * short_options = "Bb:cdefhkrtvVz";
+        {       'j', required_argument, "jobs" },
 #endif
-
-    static struct option long_options[] = { { "encode", no_argument, 0, 'e' },
-                                            { "decode", no_argument, 0, 'd' },
-                                            { "test", no_argument, 0, 't' },
-                                            { "stdout", no_argument, 0, 'c' },
-                                            { "force", no_argument, 0, 'f' },
-											{ "recover", no_argument, 0, 'r' },
-                                            { "help", no_argument, 0, 'h' },
-                                            { "keep", no_argument, 0, 'k' },
-                                            { "version", no_argument, 0, 'V' },
-                                            { "verbose", no_argument, 0, 'v' },
-                                            { "block", required_argument, 0, 'b' },
-                                            { "batch", no_argument, 0, 'B' },
-#ifdef PTHREAD
-                                            { "jobs", required_argument, 0, 'j' },
-#endif
-                                            { 0, 0, 0, 0 } };
-
-    while (1) {
-        int option_index = 0;
-        int c = getopt_long(argc, argv, short_options, long_options, &option_index);
-        if (c == -1) break;
-
-        switch (c) {
-            case '?':
-                fprintf(stderr, "Try 'bzip3 --help' for more information.\n");
-                return 1;
-            case 'e':
-            case 'z':
-                mode = MODE_ENCODE;
-                break;
-            case 'd':
-                mode = MODE_DECODE;
-                break;
-			case 'r':
-                mode = MODE_RECOVER;
-                break;
-            case 't':
-                mode = MODE_TEST;
-                break;
-            case 'c':
-                force_stdstreams = 1;
-                break;
-            case 'f':
-                force = 1;
-                break;
-            case 'k':
-                break;
-            case 'h':
-                help();
-                return 0;
-            case 'V':
-                version();
-                return 0;
-            case 'B':
-                batch = 1;
-                break;
-            case 'v':
-                verbose = 1;
-                break;
+        {         0, no_argument,       NULL }
+    };
+    yarg_settings settings = {
+        .dash_dash = true,
+        .style = YARG_STYLE_UNIX,
+    };
+    yarg_result * res = yarg_parse(argc, argv, opt, settings);
+    if (res->error) {
+        fputs(res->error, stderr);
+        fputs("Try 'bzip3 --help' for more information.\n", stderr);
+        return 1;
+    }
+    // `res' is not freed later on as it has the approximate lifetime
+    // equal to the lifetime of the program overall.
+    for (int i = 0; i < res->argc; i++) {
+        switch(res->args[i].opt) {
+            case 'e': case 'z': mode = MODE_ENCODE; break;
+            case 'd': mode = MODE_DECODE; break;
+            case 'r': mode = MODE_RECOVER; break;
+            case 't': mode = MODE_TEST; break;
+            case 'c': force_stdstreams = 1; break;
+            case 'f': force = 1; break;
+            case RM_OPTION: remove_input_file = 1; break;
+            case 'k': break;
+            case 'h': help(); return 0;
+            case 'V': version(); return 0;
+            case 'B': batch = 1; break;
+            case 'v': verbose = 1; break;
             case 'b':
-                if (!is_numeric(optarg)) {
-                    fprintf(stderr, "bzip3: invalid block size: %s\n", optarg);
+                if (!is_numeric(res->args[i].arg)) {
+                    fprintf(stderr, "bzip3: invalid block size: %s\n", res->args[i].arg);
                     return 1;
                 }
-                block_size = MiB(atoi(optarg));
+                block_size = MiB(atoi(res->args[i].arg));
                 break;
 #ifdef PTHREAD
             case 'j':
-                if (!is_numeric(optarg)) {
-                    fprintf(stderr, "bzip3: invalid amount of jobs: %s\n", optarg);
+                if (!is_numeric(res->args[i].arg)) {
+                    fprintf(stderr, "bzip3: invalid amount of jobs: %s\n", res->args[i].arg);
                     return 1;
                 }
-                workers = atoi(optarg);
+                workers = atoi(res->args[i].arg);
                 break;
 #endif
         }
@@ -641,8 +637,8 @@ int main(int argc, char * argv[]) {
         switch (mode) {
             case MODE_ENCODE:
                 /* Encode each of the files. */
-                while (optind < argc) {
-                    char * arg = argv[optind++];
+                for (int i = 0; i < res->pos_argc; i++) {
+                    char * arg = res->pos_args[i];
 
                     FILE * input_des = open_input(arg);
                     char * output_name;
@@ -660,13 +656,16 @@ int main(int argc, char * argv[]) {
                     fclose(input_des);
                     close_out_file(output_des);
                     if (!force_stdstreams) free(output_name);
+                    if (remove_input_file) {
+                        remove_in_file(arg, output_des);
+                    }
                 }
                 break;
-			case MODE_RECOVER:
+            case MODE_RECOVER:
             case MODE_DECODE:
                 /* Decode each of the files. */
-                while (optind < argc) {
-                    char * arg = argv[optind++];
+                for (int i = 0; i < res->pos_argc; i++) {
+                    char * arg = res->pos_args[i];
 
                     FILE * input_des = open_input(arg);
                     char * output_name;
@@ -689,12 +688,15 @@ int main(int argc, char * argv[]) {
                     fclose(input_des);
                     close_out_file(output_des);
                     if (!force_stdstreams) free(output_name);
+                    if (remove_input_file) {
+                        remove_in_file(arg, output_des);
+                    }
                 }
                 break;
             case MODE_TEST:
                 /* Test each of the files. */
-                while (optind < argc) {
-                    char * arg = argv[optind++];
+                for (int i = 0; i < res->pos_argc; i++) {
+                    char * arg = res->pos_args[i];
 
                     FILE * input_des = open_input(arg);
                     process(input_des, NULL, mode, block_size, workers, verbose, arg);
@@ -711,9 +713,8 @@ int main(int argc, char * argv[]) {
         return 0;
     }
 
-    while (optind < argc) {
-        // Positional argument. Likely a file name.
-        char * arg = argv[optind++];
+    for (int i = 0; i < res->pos_argc; i++) {
+        char * arg = res->pos_args[i];
 
         if (f1 != NULL && f2 != NULL) {
             fprintf(stderr, "Error: too many files specified.\n");
@@ -773,8 +774,8 @@ int main(int argc, char * argv[]) {
 
     FILE *input_des = NULL, *output_des = NULL;
 
-    output_des = mode != MODE_TEST ? open_output(output, force) : NULL;
     input_des = open_input(input);
+    output_des = mode != MODE_TEST ? open_output(output, force) : NULL;
 
     if (output != f2) free(output);
 
@@ -786,6 +787,8 @@ int main(int argc, char * argv[]) {
         fprintf(stderr, "Error: Failed on fclose(stdout): %s\n", strerror(errno));
         return 1;
     }
-
+    if (remove_input_file) {
+        remove_in_file(input, output_des);
+    }
     return r;
-}
+}
\ No newline at end of file
